####################
# version 2.0
####################


"""
FASTA I/O


shortcuts:
fasta(filename):
    -filename starts with > : output mode can't be piped

    -else: input mode | can be piped
Example:
fasta('fasta.fa') | fasta('>2.fa')

classes:
    Writer
    FastaWriter
    Reader
    FastaReader
    
methods:
    fasta
    

"""
import re
import gzip
import sys

from Core.Types.pipe import Output, Input
import Core.UI.cli as cli
import Core.utils
import Core.types
#import this for fasta C reader
#import FASTA.fasta

class Writer(Output):
    """
    Generic writer class
    """
    LINE_WIDTH = 80 #line width for output

    def __init__(self):
        Output.__init__(self)

    def append(self, title, dna=None):
        """Append a sequence to file
        @arg title:    title string
        @arg dna:      dna string
        
        implement this method
        """
        pass
            
    def action(self, input):
        #for pipe
        title, dna = input
        self.append(title, dna)


         
class FastaWriter(Writer):
    """
    Fasta class to write in FASTA format
    """
    LINE_WIDTH = 80 #line width for output

    def __init__(self, f):
        """
        @arg f: file
        """
        Writer.__init__(self)
        self.out = f

    def append(self, title, dna=None):
        """Append a sequence to FASTA file
        @arg title:    title string
        @arg dna:      dna string

        """

        self.out.write('>%s\n' % title)
        for i in range(0, len(dna), self.__class__.LINE_WIDTH):
            self.out.write('%s\n' % dna[i:i+self.__class__.LINE_WIDTH])
        self.out.flush()


class Reader(Input):
    """Generic class to read FASTA with builtin iterator
    """

    def __init__(self):
        """
        """
        Input.__init__(self)

    def count(self):
        pass

    def read(self, n=1):
        """
        Read n (or less if eof) sequences from fasta file default=1
        @arg n: number of sequences to read [default=1]

        @return: list of tuple (title, sequence)
        """
        l = []
        count = 0
        for id, sequence in self:
            if count >= n:
                break
            count += 1
            l.append( (id, sequence) )
        return l

    def next(self, decode=False):
        """
        Read next sequence in file

        @param decode:  if True decode title
        @return (title, sequence) 
        """
        raise NotImplementedError

class FastaReader(Reader):
    """
    Read a fasta file. Iterator that give the header and the dna of one sequence
    each time it is called. ( ex: for header, dna in FastaReader(file): )
    
    Reader -- a Reader object
    """
    def __init__(self, f):
        Reader.__init__(self)
        self.f = f
        self.last = ''
        self.sequence = []

    def next(self):
        self.item = False
        self.eof  = False

        while 1:
            if self.last:
                self.l = self.last
            else:
                line = self.f.readline()
                if line == '':
                    self.eof = True
                    break
                else:
                    self.l = line.strip()
            self.last = ''
            if self.l == '':
                continue
            elif self.l[0] == '#':
                continue
            elif self.l[0] == '>':
                if not self.item:
                    self.item  = True
                    self.sequence = []
                    self.title = self.l[0:]
                else:
                    self.last = self.l
                    break
            else:
                self.sequence.append(self.l)

        if self.item:
            return self.title, re.sub(r'[^ACGTUNacgtun]', '', ''.join(self.sequence))
        if self.eof:
            raise StopIteration

def fasta(filename):
    """FASTA Reader/Writer 
    filename -- ('>out', 'in')

    if filename prefix is gz : use GZIP compression mode
    """
    
    # compression mode
    
    if Core.types.string.suffix(filename) == 'gz':
        fopen = gzip.open
    else:
        fopen = open
     

    try:
       # read write of dict mode
       if filename.startswith('>'):
           filename = filename[1:]
           return FastaWriter(fopen(filename, 'w'))

       #elif filename.startswith('[]'):
       #    filename = filename[2:]
       #    return FastaDict(fopen(filename, 'r'))
       else:
           return FastaReader(fopen(filename, 'r'))
    except IOError:
	  cli.error("Error in opening file "+filename)
	       

        
        
