####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Core.Persistence.op
# DESC Object Persistency
#                                                                              #
################################################################################
"""
:NAME:
Core.Persistence.op

:DESC:
Thread safe Object Persistency
(Not Process safe)
"""
# IMPORTS
#-------------------------------------------------------------------------------
import shelve
import cPickle as pickle
import threading
import os
import random
import sys

import Core.IO.cache as cache
import Core.Types.string
import Core.UI.cli as cli


#-------------------------------------------------------------------------------
# IMPORTS

# OBJECTS
#-------------------------------------------------------------------------------

class I_P(object):
    """
    Minimal Persitence interface
    All Persistence class must implement at least this!

    index is stored in special key !
    info keys begin with an underscore _
    meta-info key start with %

    """
    def __getitem__(self, key):
        raise NotImplementedError

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default    

    def get_list(self, keys, default=None):
        """Search for list of key
        keys -- list of keys
        return list of id/None 
        """
        return [self.get(key, default) for key in keys]
        
    def keys(self):
        raise NotImplementedError

    def values(self):
        return [self[k] for k in self.keys()]

    def __setitem__(self, key, value):
        raise NotImplementedError
    
    def __delitem__(self, key):
        raise NotImplementedError
    
    def __repr__(self):
        return 'Persitence Object default repr'



########################################
#                                      #
#      Simple Shelf   
#                                      #
########################################
class Shelf(I_P):
    """
    Simple Thread safe Shelf
    example
    p = Shelf('database')
    """
    def __init__(self, filename, mode='r'):
        """
        filename -- filename without suffix .db
        mode     -- c r w
        """
        try:
            if mode == 'r' and not os.path.isfile(filename):
                cli.error(msg='Can not open file %s' % filename)
            self.mode = mode
            self.filename = filename
            self.lock = threading.Lock()
            self.d = shelve.open(self.filename, flag=mode, protocol=2, writeback=False)
        except:
            cli.error('The file '+filename+ " is not a dump file")
            sys.exit(1)
    
    def __getitem__(self, key):
        self.lock.acquire()
        try:
            r = self.d[key]
        finally:
            self.lock.release()

        return r

    def __setitem__(self, key, value):
        self.lock.acquire()
        try:
            self.d[key] = value
            self.d.sync()
        finally:
            self.lock.release()

    def __delitem__(self, key):
        self.lock.acquire()
        try:
            self.d.__delitem__(key)
        finally:
            self.lock.release()


    def keys(self):
        return self.d.keys()

    def __del__(self):
        self.d.close()

    def close(self):
        self.d.close()
        


#def CachedP(filename, maxSize=10):
#        return cache.ObjectCache(P(filename), maxSize=maxSize)    


class P(Shelf):
    """
    Indexed shelf
    an index is a dict key: key value: data shelf key
    
    index is stored in special key !
    info keys begin with an underscore _
    meta-info key start with %
    """

    
    def __getitem__(self, key):
        #try direct access
        try:
            return super(P, self).__getitem__(key)
        except KeyError:
            idx = super(P, self).__getitem__('!')
            return super(P, self).__getitem__(idx[key])    
    
    def __setitem__(self, key, value):
        super(P, self).__setitem__(key, value)
        #invalidate keys
        super(P, self).__setitem__('_keys', None)

    def __delitem__(self, key):
        super(P, self).__delitem__(key)
        #invalidate keys
        super(P, self).__setitem__('_keys', None)


    def keys(self):
        k = self.get('_keys', None)
        if k == None:
            k = [k for k in super(P, self).keys() if str(k)[0] not in ['_', '!', '%'] ]
            if self.mode != 'r': 
                super(P, self).__setitem__('_keys', k)
        return k
   
    def index_setitem(self, idx_key, key):
        
        idx = self.get('!', {})
        idx[idx_key] = key
        super(P, self).__setitem__('!', idx)


    def index_del(self):
        super(P, self).__setitem__('!', {})
        super(P, self).__setitem__('_keys', None)

    def meta_setitem(self, key, value):
        super(P, self).__setitem__('%'+key, value)

    def meta_getitem(self, key):
        super(P, self).__gettitem__('%'+key)

    def meta_delitem(self, key):
        super(P, self).__delitem__('%'+key)
   
    def index_update(self, d):
        idx = self.get('!', {})
        idx.update(d)
        super(P, self).__setitem__('!', idx)
    
    def get_idx(self):
        return super(P, self).get('!', {})

    def list(self, start):
        return [k for k in super(P, self).keys() if  str(k).startswith(start)]

    def get_regex(self, regex):
        return [self[k] for k in self.keys() if  re.match(regex, str(k))]
    

    def __repr__(self):
        return self.get('_info', 'P shelf')


class MultiP(I_P):
    """
    readonly access
    """
    def __init__(self, l=[]):
        self.PL = []
        if type(l) is str:
            l = [ l ]
        for i in l:
            self.add(i)

    def add(self, filename):
        self.PL.append(P(filename))

    def add_P(self, p):
        self.PL.append(p)

    def __getitem__(self, key):
        for p in self.PL:
            try:
                return p[key]
            except KeyError:
                continue
        raise KeyError

        
    def keys(self):
        k = []
        for P in self.PL:
            k.extend(P.keys())
        return k

    def __setitem__(self, key, value):
        raise NotImplementedError
    
    def __delitem__(self, key):
        raise NotImplementedError
    
    def __repr__(self):
        return 'Multi Persitence Object default repr'




class CachedP(P):
    """
    Cache object in memory
    dict interface
    """
    def __init__(self, filename, maxSize=0):
        """
        p -- Persistent dict
        """
        super(CachedP, self).__init__(filename, mode='r')
        self.maxSize = maxSize
        self.m = {} #memory cache
        
    def __getitem__(self, key):
        if self.m.has_key(key):
            #print "get from cache"
            return self.m[key]
        else:
            o = super(CachedP, self).__getitem__(key)
            self.m[key] = o
            if len(self.m) > self.maxSize:
                k = self.m.keys()[random.randint(0, len(self.m)-1)]
                del(self.m[k])
            return o
            
    def has_key(self, key):
        if self.m.has_key(key):
            return True
        return super(CachedP, self).has_key(key)
    


########################################
#                                      #
#       Hierarchical ODB
#                                      #
########################################
DEFAULT_ROOT = 'P.db'
DEFAULT_DB_FILENAME = 'db'

class HP(I_P):
    """
    Hierarchical Persitence

    hierarchical object database with
      - unix path  like style
      - dict interface to get/set objects

    Example:
      p = Persistence(FileBackend('Test.db'))
      p['list/item1'] = 'item1 value'
      p['list/item2'] = 'item2 value'
      p.ls('list/?')
      OUT: ['item1', 'item2']
      
    ls:
    'x/y' 'x/y/?' ...
    special keys (end of path):
         '?': list of leaf
         '#': list of node
    """
    PATH_SEPARATOR = '/'
    
    def __init__(self, backend):
        self.backend = backend
        self.root    = []
        
    def __cut_key(self, key):
        l = key.split(HP.PATH_SEPARATOR)
        return self.root + l[:-1], l[-1]
    
    def ls(self, key):
        """List keys
        Usage
          -if path ends whith *: return list of node and leaf
          -if path ends whith ?: return list of leaf
          -if path ends whith #: return list of node

        Return: list of keys
        """

        path, key = self.__cut_key(key)

        if key == '*':
            return self.backend.ls(path)
        elif key == '?':
            return self.backend.ls(path)[1]
        elif key == '#':
            return self.backend.ls(path)[0]

    keys = ls

    def __getitem__(self, key):
        """
        Option
        -if path ends whith ?: return list of objects
        """
        path, key = self.__cut_key(key)    
        if key == '?':
            return [self.backend.getitem(path, k) for  k in self.backend.ls(path)[1]]
        else:
            return self.backend.getitem(path, key)

    def get_plus(self, key):
        """Simple Regex getitem
        only allow * at end of key
        Return all objects that math key regex 
        """
        path, key = self.__cut_key(key)
        return [self.backend.getitem(path, k) for  k in self.backend.ls(path)[1] if k.startswith(key.replace('*', ''))]

    def __setitem__(self, key, value):
        path, key = self.__cut_key(key)
        return self.backend.setitem(path, key, value)
    
    def __delitem__(self, key):
        path, key = self.__cut_key(key)    
        self.backend.delitem(path, key)
        
    def chroot(self, path):
        path, key = self.__cut_key(path)
        if not key:
            self.root = []
        else:
            self.root =  key.split(HP.PATH_SEPARATOR)
    
    def __repr__(self):
        return str(self.ls('*'))

########################################
#                                      #
#      Hierarchical ODB Backend 
#                                      #
########################################
class I_Backend:
    """Backend Interface
    
    """
    def getitem(self, path, key):
        raise NotImplementedError

    def setitem(self, path, key, value):
        raise NotImplementedError

    def delitem(self, path, key):
        raise NotImplementedError

    def ls(self, path):
        """
        return tuple (node_list, leaf_list)
        """
        raise NotImplementedError

class ShelfBackend(I_Backend):
    """Simple Shelf Backend 
    """
    def __init__(self, root=DEFAULT_ROOT, db=DEFAULT_DB_FILENAME):
        self.root = root
        self.db   = db  
        self.lock = threading.Lock()
    
    def getitem(self, path, key):
        self.lock.acquire()
        try:
            d = shelve.open(join([self.root] + path, self.db))
            r = d[key]
        finally:
            self.lock.release()
        return r

    def setitem(self, path, key, value):
        self.lock.acquire()
        try:
            create_dirs(join([self.root] + path))
            d = shelve.open(join([self.root] + path, self.db))
            d[key] = value
            d.close()
        finally:
            self.lock.release()

    
    def delitem(self, path, key):
        if key:
            d = shelve.open(join([self.root] + path, self.db))
            del d[key]
            d.close()
        else: #remove dir
            remove_dirs(join([self.root] + path))

    def ls(self, path):
        if not os.path.exists(join([self.root] + path)):
            return [], []
        nodes = [i for i in os.listdir(join([self.root] + path)) if os.path.isdir(join([self.root] + path, i))]
        d = shelve.open(join([self.root] + path, self.db))
        leafs = d.keys()
        return nodes, leafs

class FileBackend(I_Backend):
    """
    File Backend 
    """
    def __init__(self, root=DEFAULT_ROOT):
        self.root = root
        self.lock = threading.Lock()
    
    def getitem(self, path, key):
        self.lock.acquire()
        try:
            r = oload(join([self.root] + path, key))
        except:
            self.lock.release()
            print 'Can not find key [%s] %s' % (key, path)
            raise
            raise KeyError, 'Can not find key [%s] %s' % (key, path)

        self.lock.release()
        return r

    def setitem(self, path, key, value):
        self.lock.acquire()
        try:
            create_dirs(join([self.root] + path))
            osave(value, join([self.root] + path, key))
        finally:
            self.lock.release()
    
    def delitem(self, path, key):
        if key:
            os.remove(join(path, key))
        else: #remove dir
            remove_dirs(join([self.root] + path))

    def ls(self, path):
        if not os.path.exists(join([self.root] + path)):
            return [], []
        leafs = [i for i in os.listdir(join([self.root] + path)) if os.path.isfile(join([self.root] + path, i)) and not i.startswith('.')]
        nodes = [i for i in os.listdir(join([self.root] + path)) if os.path.isdir(join([self.root] + path, i)) and not i.startswith('.')]
        return nodes, leafs




#-------------------------------------------------------------------------------
# OBJECTS

# CODE
#-------------------------------------------------------------------------------

def get_list(keys, databases, removeEmpty=1, default=None):
    r = []
    assert(type(databases) == list or type(databases) == tuple)

    for j in databases:
       
        for k in keys:
            
            x = j.get(k, default)          
            if not(not x and removeEmpty):
                r.append(x)

    return r


def P_from_dir(dir):
    db = {}
    for f in os.listdir(dir):
        db[Core.Types.string.prefix(f)] = P(os.path.join(dir, f))
    return db

def o_from_dir(dir):
    db = {}
    for f in os.listdir(dir):
        db[Core.Types.string.prefix(f)] = oload(os.path.join(dir, f))
    return db



def osave(obj, filename):
    """
    Save object to file
    """
    f=open(filename, 'wb')
    pickle.dump(obj,f , 2)
    f.close()

def oload(filename):
    """
    Load Object from file
    """
    try:
        f=open(filename,'rb') 
        r=pickle.load(f)
        f.close()
        return r
    except:
        cli.warning("Can not load object from file '%s'" % filename)
        raise
    
    
def join(path, key=None):
    if not key:
        return os.path.sep.join(path)
    else:
        return os.path.sep.join(path + [key])

def create_dirs(path):
    """
    check and create dirs corresponding to list path[-1]
    
    l -- path in list format
    """
    if not os.path.exists(os.path.join(path)):
        os.makedirs(os.path.join(path))

def remove_dirs(path):
    """
    Remove dir specified by path[-1] and all children files
    
    l -- path in list format    
    """
    d = os.path.join(path)
    os.system('rm -rf ' + d)

#-------------------------------------------------------------------------------
# CODE

