####################
# version 2.0
####################


"""
Interface for building and scanning a background model

classes:
    BGDB
    BG
    Pataud
    
methods:
    load_bg
    build_BG
    

"""

from numpy import *

from math import *
import os
import sys
import glob
import platform
import Core.UI.cli as cli
import Core.Types.list
import Core.Types.index
import Core.Types.interval
import Core.Types.list
import Core.Persistence.op as op
import Core.IO.cache as cache

import Bio.sequence


import score_dist
import TFBSscanner




# IMPORT BINARY MODULE !! EDIT IF NEEDED
try:
    import Pataud.cpataud as cpataud
except ImportError:
    try:
        if sys.platform == 'darwin':
            
            if sys.byteorder == 'little':
                archi=platform.architecture()
                if archi[0] == '64bit':
                    import Pataud.MacIntel64.cpataud as cpataud
                else:
                    import Pataud.MacIntel32.cpataud as cpataud
            else:
                import Pataud.Macppc.cpataud as cpataud
        elif sys.platform == 'linux2':
            archi=platform.architecture()
            if archi[0] == '64bit':
                import Pataud.Linux64.cpataud as cpataud
            else:
                import Pataud.Linux32.cpataud as cpataud
        ## a tester
        elif sys.platform == 'win32':
            import Pataud.Windows.cpataud as cpataud
        else:
            raise ImportError, 'Can not load Pataud binary module. Try to compile it using Pyrex '
            sys.exit(1)
    except:
        raise ImportError, 'Can not load Pataud binary module. Try to compile it using Pyrex '
        sys.exit(1)
#

DUMP_DATA_TYPE = 1

def load_bg(file=None, files=None, dir=None, loc=None):
    """
    load a background model from one or several files
    
    file   -- a file with a background model
    
    files  -- a list of files with background models
    
    dir    -- a directory with files with background models
    
    return -- an object containing background models
    """
    bg = BGDB()

    if file:
        bg.add_dump(file,loc)
    elif files != None:
        for file in files:
            bg.add_dump(file,loc)
    elif os.path.isdir(dir):
        for f in os.listdir(dir):
            bg.add_dump(os.path.join(dir, f),loc)            

    bg.index()
    return bg


class BGDB(TFBSscanner.I_BGDB):
    """
    Indexed database of TFM-Explorer Background Model
    
    Methods
    -------
    add_dump -- add TFM-Explorer BG file to the database
    index    -- index the database
    get      -- retreive a BG object given a sequence and a matrix


    Internal Structure
    ------------------
    .data store a list of BG model (dataIdx referes to an index in this list)

    object: tuple (datatype, dataIdx, matrixId)

    .idx store an index of BG model : association between (matrix ID, Organism, GC range) and BG model
    
    """


    def __init__(self):

        self.scanner = Pataud()
        self.idx = Core.Types.index.SimpleIndex(('matrixId', 'organism', 'gcRange'))
        self.data = []

    def obj2idx(self, o):
        datatype, dataIdx, matrixId = o
        p = self.data[dataIdx]
        bg = p[matrixId]
        return (bg.matrix.id, bg.organism, bg.gcRange)

    def add_dump(self, dumpfilename,loc):
        """
        Add a background model to a list. Check before that it is correct.
        
        dumpfilename -- the dump file with the background model to add
        
        loc          -- the location of the sequences that should correspond to the background model
        
        WARNING each dump file MUST use the same TFBSscanner
        
        """
        # add dumpfilename
        if not os.path.exists(dumpfilename):
            print >> sys.stderr, 'Warning %s not found' % dumpfilename

        p = op.CachedP(dumpfilename, 20)
        
        if p.get('_info').find('Pataud') != -1:
            if loc[0]< p.get('_location')[0]  or loc[1] > p.get('_location')[1] :
                cli.warning('file '+dumpfilename+' has incorrect location. Cannot be used as a background model for these sequences. Ignored.')                
            else:
                dataIdx = len(self.data)
                self.data.append(p)
                keys = p.keys()
                ol = zip([DUMP_DATA_TYPE]*len(keys), [dataIdx]*len(keys), keys)
                self.idx.add_list(ol)
        else:
            cli.warning('file '+ dumpfilename+' is not a background model. Ignored.')

    def index(self):
        #build index

        self.idx.build(self.obj2idx)
        self.organisms = self.idx.list_values('organism')
        self.gcRange = {}
        for org in self.organisms:
            self.gcRange[org] = self.idx.list_values('gcRange', ('organism', org))

    def get(self, sequence, matrix):
        """
        Return background model data for a matrix and a sequence
        
        sequence -- an object Sequence
        
        matrix   -- a pwm 
        """
        matrixId = matrix.id
        organism = sequence.organism
        gc = sequence.gc()

        if organism not in self.gcRange.keys() :
            cli.error("An organism cannot be linked with a background model: "+ organism)
            sys.exit(1)
        else:
            gcRange = Core.Types.interval.find(gc, self.gcRange[organism])
            l = list(self.idx.get( (matrixId, organism, gcRange) ))        
            datatype, dataIdx, matrixId = Core.Types.list.get(l, 0, (None, None, None))
            if datatype == DUMP_DATA_TYPE:
                p = self.data[dataIdx]
        
                return p[matrixId]
            else:
                return None
                
    def __repr__(self):
        return '\n'.join([i.filename for i in self.data])


def build_BG(sequences, matrices, priori_p, filename, overlap=True, gcRange=None ):
    """
    constructs a background model
    
    sequences -- list of sequences
    
    matrices  -- list of matrices
    
    priori_p  -- priori probability
    
    filename  -- name of the outputted file
    
    """
    p = op.P(filename, mode='c')

    p['_info'] = 'Pataud BG DB [%.4d:%.4d] p=%f' % (sequences[0].location[0], sequences[0].location[1], priori_p)
    p['_location'] = sequences[0].location
    p['_priori_p'] = priori_p
    p['_numberOfSequences'] = len(sequences)
    p['_TFBSscanner'] = Pataud()

    
    info = cli.Info(len(matrices), 1)

    for matrix in matrices:
        bg = BG(sequences, matrix, priori_p, overlap, gcRange=gcRange)
        p[matrix.id] = bg
        info('Building pataud BG %s ' % (matrix.id))    
    p.keys()    
class BG(TFBSscanner.I_BG):
    """
    TFM-Explorer Background Model
    """

    def __init__(self, sequences, matrix, p=0.005, overlap=True, gcRange=None):
        self.matrix   = matrix
        self.location = sequences[0].location
        self.organism = sequences[0].organism
        self.overlap  = overlap 

        #simple strand priori
        self.priori        = p / 2.0
        self.gcRange = gcRange
        self.__build(self.matrix, sequences)
        self.compute_p(self.matrix, sequences, Pataud())

    def __build(self, matrix, sequences, windowWidth=400, overlap=False):
        """
        Build Bg for both strand

        matrix      -- one matrix
        
        sequences   -- list of aligned sequences
    
        p           -- reach this pvalue
        
        windowWidth -- use this window width to compute distribution
    
        return (acgt_table, table_of_score_threshold)  #for each position of the sequences alignement
        """
	#windowWidth= 1000
        #TYPE = 0 #-1: obs dist, 0: MM0, 1: MM1
        SMOOTHING_WINDOW_WIDTH = windowWidth
        EPSILON = 0.1
       # self.type = TYPE

        #priori probability
        priori = self.priori

        #Lengths
        S = len(sequences)
        L = len(sequences[0])
        W = windowWidth
        M = len(matrix)

         # step
        STEP = W

        # local frequencies
        ACGT = {}
        ACGT_x={}
        ACGT_t = [0.25]*4
        AT = (ACGT_t[0] + ACGT_t[3]) / 2.0
        GC = (ACGT_t[1] + ACGT_t[2]) / 2.0
        ACGT_t[0], ACGT_t[3] = AT, AT
        ACGT_t[1], ACGT_t[2] = GC, GC
        
        for I in range(0, L-W+1-M+1, STEP):
            ACGT[I] = Bio.sequence.ACGT_freq(sequences, I, I+W)
            ACGT_x[I]=ACGT_t



        #resTab
        threshold = zeros( L, 'f')
        #p         = zeros( L, 'f')
        pACGT = zeros( (L, 4), 'f')

        self.epsilon = EPSILON
        assert(L >= W + M)
        #info = cli.Info((L-W+1-M+1)/float(STEP), 1)
        #for each window
        for I in range(0, L-W+1-M+1, STEP):
            lACGT = ACGT[I]
 
            # compute the PWM from the frequency matrix taking into
            # account local frequencies
            pwm = Bio.pwm.PWM(matrix, lACGT)
            # m is the rounded matrix
            m = zeros( (len(pwm.w), 4) )
            #epsilon = self.epsilon 
            H0 = {0:1.0}          
            for pos in range(len(m)):
                H1   = {}
                for letter in range(4):
                    m[pos,letter] = int(round((pwm.w[pos,letter] / self.epsilon)))
                    for sc in H0.keys():
                        S = sc + m[pos,letter]
                        H1[S] = H1.get(S, 0) + \
                            H0[sc] * lACGT[letter]
                H0 = H1                        
            d = H0
                    
            # score distribution computation
            #d=score_dist.score_distribution(m,lACGT)
            
            # threshold computation
            thr = score_dist.pvalue2score(d, self.priori) * self.epsilon - self.epsilon

            #fill tables
            pACGT[I:I+STEP,:] = ACGT_x[I]
            threshold[I:I+STEP] = thr

        #fill ends of tables
        pACGT[I+STEP:,:] = ACGT_x[I]
        threshold[I+STEP:] = thr


        #smooth tab
        threshold = Core.Types.tab.smooth(threshold, SMOOTHING_WINDOW_WIDTH)
        for b in range(4):
            pACGT[:,b] = Core.Types.tab.smooth(pACGT[:,b], SMOOTHING_WINDOW_WIDTH)

        #set attribute
        #self.p         = p 
        self.pACGT     = pACGT
        self.threshold = threshold
        

class Pataud(TFBSscanner.I_TFBSscanner):
    """
    Interface for scanning a background model
    """

    def __init__(self, bg=None):
        self.bg      = bg

    def scan(self, sequences, matrix, bg):
        """
        scan a background model with sequences and matrix
        
        Use a C library
        """
        #init c object
        csequences = cpataud.Sequences(sequences)
        cmatrix = cpataud.Matrix(matrix)
        cbg = cpataud.Bg(bg)
        overlap = bg.overlap and 1 or 0
        r = cpataud.score_sequences_bg(cmatrix, csequences, cbg, overlap)
        #r = score_sequences_bg(matrix, sequences, bg, overlap)

        return r


if __name__ == '__main__':
    pass
